package com.io7m.glowmaps;

import java.awt.image.BufferedImage;
import java.awt.image.ColorConvertOp;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import javax.imageio.ImageIO;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL21;
import org.lwjgl.opengl.GL30;

public final class Utilities
{
  private static ByteBuffer readFile(
    final String name)
    throws IOException
  {
    final File file = new File(name);
    final BufferedInputStream stream =
      new BufferedInputStream(new FileInputStream(file));
    final ByteBuffer buffer = ByteBuffer.allocateDirect((int) file.length());

    for (;;) {
      final int ch = stream.read();
      if (ch == -1) {
        break;
      }
      buffer.put((byte) ch);
    }

    buffer.rewind();
    buffer.order(ByteOrder.nativeOrder());

    stream.close();
    return buffer;
  }

  static int createShader(
    final String vertex_shader,
    final String fragment_shader)
    throws IOException
  {
    int status;

    final int program = GL20.glCreateProgram();
    final int vs = GL20.glCreateShader(GL20.GL_VERTEX_SHADER);
    final int fs = GL20.glCreateShader(GL20.GL_FRAGMENT_SHADER);
    Utilities.checkGL();

    final ByteBuffer vs_buffer = Utilities.readFile(vertex_shader);
    GL20.glShaderSource(vs, vs_buffer);
    GL20.glCompileShader(vs);
    status = GL20.glGetShader(vs, GL20.GL_COMPILE_STATUS);
    if (status != GL11.GL_TRUE) {
      throw new RuntimeException("shader compilation failure: "
        + GL20.glGetShaderInfoLog(vs, 8192));
    }

    final ByteBuffer fs_buffer = Utilities.readFile(fragment_shader);
    GL20.glShaderSource(fs, fs_buffer);
    GL20.glCompileShader(fs);
    status = GL20.glGetShader(fs, GL20.GL_COMPILE_STATUS);
    if (status != GL11.GL_TRUE) {
      throw new RuntimeException("shader compilation failure: "
        + GL20.glGetShaderInfoLog(fs, 8192));
    }

    GL20.glAttachShader(program, vs);
    GL20.glAttachShader(program, fs);
    GL20.glLinkProgram(program);
    status = GL20.glGetProgram(program, GL20.GL_LINK_STATUS);
    if (status != GL11.GL_TRUE) {
      throw new RuntimeException("shader link failure: "
        + GL20.glGetProgramInfoLog(program, 8192));
    }

    return program;
  }

  static void checkGL()
  {
    final int e = GL11.glGetError();
    if (e != GL11.GL_NO_ERROR) {
      throw new RuntimeException("OpenGL error: code " + e);
    }
  }

  static int createEmptyTexture(
    final int width,
    final int height)
  {
    final int b = GL15.glGenBuffers();
    Utilities.checkGL();

    final int size = width * height * 4;
    GL15.glBindBuffer(GL21.GL_PIXEL_UNPACK_BUFFER, b);
    GL15.glBufferData(GL21.GL_PIXEL_UNPACK_BUFFER, size, GL15.GL_STATIC_DRAW);
    GL15.glBindBuffer(GL21.GL_PIXEL_UNPACK_BUFFER, 0);
    Utilities.checkGL();

    final int t = GL11.glGenTextures();
    Utilities.checkGL();

    GL15.glBindBuffer(GL21.GL_PIXEL_UNPACK_BUFFER, b);
    GL11.glBindTexture(GL11.GL_TEXTURE_2D, t);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_MIN_FILTER,
      GL11.GL_LINEAR);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_MAG_FILTER,
      GL11.GL_LINEAR);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_WRAP_S,
      GL11.GL_REPEAT);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_WRAP_T,
      GL11.GL_REPEAT);
    GL11.glTexImage2D(
      GL11.GL_TEXTURE_2D,
      0,
      GL11.GL_RGBA,
      width,
      height,
      0,
      GL11.GL_RGBA,
      GL11.GL_UNSIGNED_BYTE,
      0L);
    GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
    GL15.glBindBuffer(GL21.GL_PIXEL_UNPACK_BUFFER, 0);
    Utilities.checkGL();

    return t;
  }

  static int createFramebuffer(
    final int texture)
  {
    final int fb = GL30.glGenFramebuffers();
    Utilities.checkGL();

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fb);
    GL30.glFramebufferTexture2D(
      GL30.GL_FRAMEBUFFER,
      GL30.GL_COLOR_ATTACHMENT0,
      GL11.GL_TEXTURE_2D,
      texture,
      0);

    final int status = GL30.glCheckFramebufferStatus(GL30.GL_FRAMEBUFFER);
    if (status != GL30.GL_FRAMEBUFFER_COMPLETE) {
      throw new RuntimeException("framebuffer incomplete: status " + status);
    }

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    return fb;
  }

  static final BufferedImage convertImageToABGR(
    final BufferedImage image)
  {
    BufferedImage converted = null;

    final int width = image.getWidth();
    final int height = image.getHeight();

    switch (image.getType()) {
      case BufferedImage.TYPE_4BYTE_ABGR:
      case BufferedImage.TYPE_4BYTE_ABGR_PRE:
      {
        converted = image;
        return converted;
      }
      case BufferedImage.TYPE_INT_ARGB_PRE:
      case BufferedImage.TYPE_INT_ARGB:
      case BufferedImage.TYPE_3BYTE_BGR:
      case BufferedImage.TYPE_BYTE_GRAY:
      case BufferedImage.TYPE_BYTE_BINARY:
      case BufferedImage.TYPE_BYTE_INDEXED:
      case BufferedImage.TYPE_USHORT_GRAY:
      case BufferedImage.TYPE_USHORT_565_RGB:
      case BufferedImage.TYPE_USHORT_555_RGB:
      case BufferedImage.TYPE_CUSTOM:
      case BufferedImage.TYPE_INT_RGB:
      case BufferedImage.TYPE_INT_BGR:
      default:
      {
        converted =
          new BufferedImage(width, height, BufferedImage.TYPE_4BYTE_ABGR);
        final ColorConvertOp op = new ColorConvertOp(null);
        op.filter(image, converted);
        return converted;
      }
    }
  }

  public static int loadTexture(
    final String string,
    final int filter)
    throws IOException
  {
    final FileInputStream stream = new FileInputStream(string);
    final BufferedImage image = ImageIO.read(stream);

    final BufferedImage converted = Utilities.convertImageToABGR(image);
    final ByteBuffer map =
      ByteBuffer.allocateDirect(image.getHeight() * image.getWidth() * 4);
    map.order(ByteOrder.nativeOrder());

    final WritableRaster raster = converted.getRaster();
    final DataBufferByte data = (DataBufferByte) raster.getDataBuffer();
    final byte[] bytes = data.getData();
    for (int index = 0; index < bytes.length; index += 4) {
      map.put(bytes[index + 3]);
      map.put(bytes[index + 2]);
      map.put(bytes[index + 1]);
      map.put(bytes[index + 0]);
    }
    map.rewind();
    stream.close();

    final int t = GL11.glGenTextures();
    Utilities.checkGL();

    GL11.glBindTexture(GL11.GL_TEXTURE_2D, t);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_MIN_FILTER,
      filter);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_MAG_FILTER,
      filter);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_WRAP_S,
      GL11.GL_REPEAT);
    GL11.glTexParameteri(
      GL11.GL_TEXTURE_2D,
      GL11.GL_TEXTURE_WRAP_T,
      GL11.GL_REPEAT);
    GL11.glTexImage2D(
      GL11.GL_TEXTURE_2D,
      0,
      GL11.GL_RGBA,
      image.getWidth(),
      image.getHeight(),
      0,
      GL11.GL_RGBA,
      GL11.GL_UNSIGNED_BYTE,
      map);
    GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
    Utilities.checkGL();

    return t;
  }

  /**
   * Render a quad of (width * height), centered at (x, y).
   */

  static void renderQuadCenteredAt(
    final double x,
    final double y,
    final double width,
    final double height)
  {
    GL11.glPushMatrix();
    {
      GL11.glTranslated(x - (width / 2), y - (height / 2), 0);
      GL11.glBegin(GL11.GL_QUADS);
      {
        GL11.glTexCoord2d(0, 0);
        GL11.glVertex3d(0, 0, 0);
        GL11.glTexCoord2d(1, 0);
        GL11.glVertex3d(width, 0, 0);
        GL11.glTexCoord2d(1, 1);
        GL11.glVertex3d(width, height, 0);
        GL11.glTexCoord2d(0, 1);
        GL11.glVertex3d(0, height, 0);
      }
      GL11.glEnd();
    }
    GL11.glPopMatrix();
    Utilities.checkGL();
  }

  /**
   * Render a quad of (width * height), with the top-left corner at (x, y).
   */

  static void renderQuadCorneredAt(
    final double x,
    final double y,
    final double width,
    final double height)
  {
    GL11.glPushMatrix();
    {
      GL11.glTranslated(x, y, 0);
      GL11.glBegin(GL11.GL_QUADS);
      {
        GL11.glTexCoord2d(0, 0);
        GL11.glVertex3d(0, 0, 0);
        GL11.glTexCoord2d(1, 0);
        GL11.glVertex3d(width, 0, 0);
        GL11.glTexCoord2d(1, 1);
        GL11.glVertex3d(width, height, 0);
        GL11.glTexCoord2d(0, 1);
        GL11.glVertex3d(0, height, 0);
      }
      GL11.glEnd();
    }
    GL11.glPopMatrix();
    Utilities.checkGL();
  }

  /**
   * Render a fullscreen quad.
   */

  static void renderScreenAlignedQuad(
    final int width,
    final int height)
  {
    GL11.glMatrixMode(GL11.GL_PROJECTION);
    GL11.glPushMatrix();
    GL11.glLoadIdentity();
    GL11.glOrtho(0, width, 0, height, 1, 100);
    Utilities.checkGL();

    GL11.glMatrixMode(GL11.GL_MODELVIEW);
    GL11.glPushMatrix();
    GL11.glLoadIdentity();
    GL11.glTranslated(0, 0, -1);
    Utilities.checkGL();

    Utilities.renderQuadCorneredAt(0, 0, width, height);

    GL11.glMatrixMode(GL11.GL_PROJECTION);
    GL11.glPopMatrix();
    GL11.glMatrixMode(GL11.GL_MODELVIEW);
    GL11.glPopMatrix();
    Utilities.checkGL();
  }
}
